from urllib.parse import quote
from typing import Iterable, Literal
import aiohttp

from .types import (
    ActionName,
    ArgumentName,
    AttributeName,
    OzmaDBActionResponse,
    OzmaDBSavedSchemas,
    OzmaDBTransactionOp,
    OzmaDBTransactionResult,
    OzmaDBViewExprResult,
    OzmaQLValue,
    OzmaQLValueArg,
    RawOzmaQLValue,
    SchemaName,
    UserViewName,
    ozmaql_value_to_raw,
)
from .errors import check_ozma_response
from .auth import OzmaAuth


def ozmaql_name_to_str(v: str) -> str:
    escaped = v.replace('"', '\\"')
    return f'"{escaped}"'


def ozmaql_value_to_str(v: OzmaQLValue) -> str:
    match v:
        case None:
            return "NULL"
        case bool():
            return "TRUE" if v else "FALSE"
        case str():
            escaped = v.replace("'", "\\'")
            return f"'{escaped}'"
        case int() | float():
            return str(v)
        # TODO: Add support for datetime and timedelta.
        case _:
            raise TypeError(f"Invalid OzmaQL value type {type(v)}")


def dict_to_attributes(d: dict[AttributeName, OzmaQLValue]) -> str:
    escaped = ", ".join(f"{ozmaql_name_to_str(k)} = {ozmaql_value_to_str(v)}" for k, v in d.items())
    return "@{ " + escaped + " }"


# TODO: Convert manual dicts to Pydantic classes.
class OzmaDBAPI:
    BASE_URL = "https://{name}.api.ozma.org"
    _api_url: str
    _auth: OzmaAuth

    def __init__(self, auth: OzmaAuth, *, api_url: str | None = None, name: str | None = None):
        self._auth = auth
        if not ((api_url is not None) ^ (name is not None)):
            raise ValueError("Either api_url or name must be set")
        if api_url is not None:
            self._api_url = api_url
        else:
            self._api_url = self.BASE_URL.format(name=name)

    async def get_named_user_view(
        self, schema: SchemaName, name: UserViewName, args: dict[ArgumentName, OzmaQLValueArg] | None = None
    ) -> OzmaDBViewExprResult:
        async with aiohttp.ClientSession() as session:
            request = {}
            if args is not None:
                request["args"] = {name: ozmaql_value_to_raw(arg) for name, arg in args}
            async with self._auth.with_auth(
                session.post,
                f"{self._api_url}/views/by_name/{quote(schema)}/{quote(name)}/entries",
                headers={"Accept": "application/json"},
                json=request,
            ) as response:
                await check_ozma_response(response)
                ret = await response.json()
                return OzmaDBViewExprResult.model_validate(ret)

    async def run_transaction(
        self,
        operations: Iterable[OzmaDBTransactionOp],
        *,
        defer_constraints: bool = False,
        force_allow_broken: bool = False,
    ) -> OzmaDBTransactionResult:
        async with aiohttp.ClientSession() as session:
            params = {
                "operations": [op.model_dump(by_alias=True) for op in operations],
                "defer_constraints": defer_constraints,
                "force_allow_broken": force_allow_broken,
            }
            print(params)
            async with self._auth.with_auth(
                session.post,
                f"{self._api_url}/transaction",
                json=params,
            ) as response:
                await check_ozma_response(response)
                ret = await response.json()
                return OzmaDBTransactionResult.model_validate(ret)

    async def run_action(
        self, schema: SchemaName, name: ActionName, args: dict[str, OzmaQLValueArg] | None = None
    ) -> OzmaDBActionResponse:
        raw_args: dict[str, RawOzmaQLValue]
        if args is None:
            raw_args = {}
        else:
            raw_args = {name: ozmaql_value_to_raw(arg) for name, arg in args}
        async with aiohttp.ClientSession() as session:
            async with self._auth.with_auth(
                session.post,
                f"{self._api_url}/actions/{quote(schema)}/{quote(name)}/run",
                json=raw_args,
            ) as response:
                await check_ozma_response(response)
                ret = await response.json()
                return OzmaDBActionResponse.model_validate(ret)

    async def save_schemas(
        self, schemas: list[str] | Literal["all"], *, skip_preloaded: bool | None = None
    ) -> OzmaDBSavedSchemas:
        async with aiohttp.ClientSession() as session:
            params: dict[str, str | list[str]] = {}
            if schemas != "all":
                params["schema"] = schemas
            if skip_preloaded is not None:
                params["skip_preloaded"] = str(skip_preloaded)
            async with self._auth.with_auth(
                session.get,
                f"{self._api_url}/layouts",
                params=params,
                headers={"Accept": "application/json"},
            ) as response:
                await check_ozma_response(response)
                ret = await response.json()
                return OzmaDBSavedSchemas.model_validate(ret)

    async def restore_schemas(
        self,
        schemas: OzmaDBSavedSchemas,
        *,
        drop_others: bool | None = None,
        force_allow_broken: bool | None = None,
    ):
        async with aiohttp.ClientSession() as session:
            params: dict[str, str | list[str]] = {}
            if drop_others is not None:
                params["drop_others"] = str(drop_others)
            if force_allow_broken is not None:
                params["force_allow_broken"] = str(force_allow_broken)
            async with self._auth.with_auth(
                session.put,
                f"{self._api_url}/layouts",
                params=params,
                json=schemas.model_dump(by_alias=True)["schemas"],
            ) as response:
                await check_ozma_response(response)



from typing import Any
import aiohttp
from pydantic import BaseModel, field_serializer, field_validator
from datetime import datetime, timedelta
import urllib.parse

from .errors import check_ozma_response
from .auth import OzmaAuth


class InstancesModel(BaseModel, extra="ignore"):
    pass


class RateLimit(InstancesModel):
    limit: int
    period: timedelta

    @field_serializer("period", mode="plain")
    @staticmethod
    def serialize_period(value: timedelta) -> int:
        return int(value.total_seconds())

    @field_validator("period", mode="before")
    @staticmethod
    def validate_period(time) -> timedelta:
        match time:
            case timedelta():
                return time
            case int() | float():
                return timedelta(seconds=time)
            case _:
                raise ValueError("Invalid period")


class InstanceRateLimits(InstancesModel):
    max_size: int | None = None
    max_request_time: timedelta | None = None
    max_users: int | None = None
    read_rate_limits_per_user: list[RateLimit] | None = None
    write_rate_limits_per_user: list[RateLimit] | None = None

    @field_serializer("max_request_time", mode="plain")
    @staticmethod
    def serialize_max_request_time(value: timedelta | None) -> int | None:
        if value is None:
            return None
        return int(value.total_seconds() * 1000)

    @field_validator("max_request_time", mode="before")
    @staticmethod
    def validate_max_request_time(time) -> timedelta | None:
        match time:
            case timedelta():
                return time
            case int() | float():
                return timedelta(seconds=time)
            case _:
                raise ValueError("Invalid max_request_time")


class Instance(InstanceRateLimits):
    name: str
    owner: str
    region: str
    created_at: datetime
    accessed_at: datetime | None = None

    enabled: bool = True
    is_template: bool = False
    hidden: bool = False

    published: bool = True
    anyone_can_read: bool = False
    disable_security: bool = False


class NewInstance(InstanceRateLimits):
    name: str
    owner: str | None = None
    group: str | None = None
    randomize_name: bool = False

    hidden: bool = False
    is_template: bool = False

    published: bool = True
    disable_security: bool = False
    anyone_can_read: bool = False


class Instances(InstancesModel):
    instances: list[Instance]


class OzmaInstancesAPI:
    _api_url = "https://instances.services.ozma.io"
    _auth: OzmaAuth

    def __init__(self, auth: OzmaAuth, *, api_url: str | None = None):
        if api_url is not None:
            self._api_url = api_url
        self._auth = auth

    async def get_instances(
        self,
        *,
        owner: str | None = None,
        enabled_only: bool | None = None,
        or_template: bool | None = None,
    ) -> Instances:
        async with aiohttp.ClientSession() as session:
            params: dict[str, Any] = {}
            if owner is not None:
                params["owner"] = owner
            if enabled_only is not None:
                params["enabled_only"] = enabled_only
            if or_template is not None:
                params["or_template"] = or_template
            async with self._auth.with_auth(session.get, f"{self._api_url}/instances", params=params) as response:
                await check_ozma_response(response)
                ret = await response.json()
                return Instances.model_validate(ret)

    async def create_instance(self, instance: NewInstance) -> Instance:
        async with aiohttp.ClientSession() as session:
            async with self._auth.with_auth(
                session.post,
                f"{self._api_url}/instances",
                # FIXME: We have a mess with the default values vs `null` values in the instances service.
                # For now, just exclude the defaults.
                json=instance.model_dump(exclude_defaults=True, by_alias=True),
            ) as response:
                await check_ozma_response(response)
                ret = await response.json()
                return Instance.model_validate(ret)

    async def delete_instance(self, name: str) -> None:
        async with aiohttp.ClientSession() as session:
            async with self._auth.with_auth(
                session.delete,
                f"{self._api_url}/instances/{urllib.parse.quote(name, safe='')}",
            ) as response:
                await check_ozma_response(response)



from dataclasses import dataclass
from enum import Enum
from typing import Literal


def ozmaql_name(name: str) -> str:
    return '"{}"'.format(name.replace('"', '""'))


def _ozmaql_string_char(char: str) -> str:
    if char == "\\":
        return "\\\\"
    elif char == "\b":
        return "\\b"
    elif char == "\f":
        return "\\f"
    elif char == "\n":
        return "\\n"
    elif char == "\r":
        return "\\r"
    elif char == "\t":
        return "\\t"
    elif char == "'":
        return "\\'"
    elif ord(char) < ord(" "):
        return "\\x{:02X}".format(ord(char))
    else:
        return char


def ozmaql_string(string: str) -> str:
    return "'{}'".format("".join((_ozmaql_string_char(char) for char in string)))


@dataclass
class OzmaQLRaw:
    raw: str


# This is a hack to allow `Literal[FUNQL_UNDEFINED]` to be used in `JsonInnerValue`.
class _OzmaQLUndefined(Enum):
    UNDEFINED = 0


FUNQL_UNDEFINED = _OzmaQLUndefined.UNDEFINED


JsonValue = None | str | int | float | bool | OzmaQLRaw | dict[str, "JsonInnerValue"] | list["JsonInnerValue"]
JsonInnerValue = JsonValue | Literal[FUNQL_UNDEFINED]


def ozmaql_json(json: JsonValue, *, pretty_offset: int | None = None, current_offset: int = 0) -> str:
    match json:
        case None:
            return "null"
        case str():
            return ozmaql_string(json)
        case int():
            return str(json)
        case float():
            return str(json)
        case bool():
            return "true" if json else "false"
        case OzmaQLRaw():
            return json.raw
        case dict():
            filtered_dict = {key: value for key, value in json.items() if value is not FUNQL_UNDEFINED}
            if len(filtered_dict) == 0:
                return "{}"
            elif pretty_offset is None:
                return (
                    "{"
                    + ", ".join(f"{ozmaql_name(key)}: {ozmaql_json(value)}" for key, value in filtered_dict.items())
                    + "}"
                )
            else:
                return (
                    "{\n"
                    + "".join(
                        f"{' ' * (current_offset + pretty_offset)}{ozmaql_name(key)}: {ozmaql_json(value, pretty_offset=pretty_offset, current_offset=current_offset + pretty_offset)},\n"
                        for key, value in filtered_dict.items()
                    )
                    + " " * current_offset
                    + "}"
                )
        case list():
            filtered_list = [value for value in json if value is not FUNQL_UNDEFINED]
            if len(filtered_list) == 0:
                return "[]"
            elif pretty_offset is None:
                return (
                    "[" + ", ".join(ozmaql_json(value, current_offset=current_offset) for value in filtered_list) + "]"
                )
            else:
                return (
                    "[\n"
                    + "".join(
                        f"{' ' * (current_offset + pretty_offset)}{ozmaql_json(value, pretty_offset=pretty_offset, current_offset=current_offset + pretty_offset)},\n"
                        for value in filtered_list
                    )
                    + " " * current_offset
                    + "]"
                )
        case _:
            raise TypeError(f"Invalid JSON value type {type(json)}")



import asyncio
from types import TracebackType
from typing import Any, Callable, Coroutine, Generator, Type
import aiohttp
import jwt
from datetime import datetime, timedelta


_LAG_MULTIPLIER = 0.9


# Needed because `asyncio` has its own Awaitable type we can't reuse.
class _AuthResponseContextManager(Coroutine[Any, Any, aiohttp.ClientResponse]):
    _coro: Coroutine[Any, Any, aiohttp.ClientResponse]
    _response: aiohttp.ClientResponse | None
    __slots__ = ("_coro", "_response")

    def __init__(self, coro: Coroutine[Any, Any, aiohttp.ClientResponse]):
        self._coro = coro

    def send(self, arg: None) -> asyncio.Future[Any]:
        return self._coro.send(arg)

    def throw(self, *args: Any, **kwargs: Any) -> asyncio.Future[Any]:
        return self._coro.throw(*args, **kwargs)

    def close(self) -> None:
        return self._coro.close()

    def __await__(self) -> Generator[Any, None, aiohttp.ClientResponse]:
        ret = self._coro.__await__()
        return ret

    async def __aenter__(self):
        self._response = await self._coro
        return self._response

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ):
        assert self._response is not None
        self._response.release()
        await self._response.wait_for_close()


class OzmaAuth:
    _api_url = "https://account.ozma.io/auth/realms/default/.well-known/openid-configuration"
    _client_id: str
    _client_secret: str
    _login: str | None = None
    _password: str | None = None
    _lock: asyncio.Lock

    _token_url: str | None = None
    _signing_algorithms: list[str] | None = None
    _jwks_client: jwt.PyJWKClient | None = None

    _access_token: str | None = None
    _decoded_access_token: dict[str, Any] | None = None
    _access_expired_at: datetime | None = None

    _refresh_token: str | None = None
    _decoded_refresh_token: dict[str, Any] | None = None
    _refresh_expired_at: datetime | None = None

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        *,
        api_url: str | None = None,
        login: str | None = None,
        password: str | None = None,
    ):
        self._client_id = client_id
        self._client_secret = client_secret
        if (login is None) ^ (password is None):
            raise ValueError("Login and password must both be set")
        if api_url is not None:
            self._api_url = api_url
        self._login = login
        self._password = password
        self._lock = asyncio.Lock()

    async def _get_configuration(self):
        async with aiohttp.ClientSession() as session:
            async with session.get(self._api_url) as response:
                response.raise_for_status()
                ret = await response.json()
                self._token_url = ret["token_endpoint"]
                self._signing_algorithms = ret["id_token_signing_alg_values_supported"]
                self._jwks_client = jwt.PyJWKClient(ret["jwks_uri"])

    async def _ensure_configuration(self):
        if self._token_url is None:
            await self._get_configuration()

    async def _parse_response(self, response: dict[str, Any]):
        assert self._signing_algorithms is not None
        assert self._jwks_client is not None
        now = datetime.now()

        access_token = response["access_token"]
        signing_key = await asyncio.to_thread(self._jwks_client.get_signing_key_from_jwt, access_token)

        def decode_jwt(token: str):
            return jwt.decode(
                token,
                signing_key.key,
                algorithms=self._signing_algorithms,
                audience="account",
            )

        decoded_access_token = decode_jwt(access_token)
        if "refresh_token" in response:
            refresh_token = response["refresh_token"]
            decoded_refresh_token = decode_jwt(refresh_token)
        else:
            refresh_token = None
            decoded_refresh_token = None

        self._access_token = access_token
        self._decoded_access_token = decoded_access_token
        self._access_expired_at = now + _LAG_MULTIPLIER * timedelta(seconds=response["expires_in"])
        if refresh_token is not None:
            self._refresh_token = refresh_token
            self._decoded_refresh_token = decoded_refresh_token
            self._refresh_expired_at = now + _LAG_MULTIPLIER * timedelta(seconds=response["refresh_expires_in"])

    async def _get_tokens(self):
        await self._ensure_configuration()
        assert self._token_url is not None

        if self._login is None:
            # Authenticate with a service account.
            data = {
                "grant_type": "client_credentials",
                "client_id": self._client_id,
                "client_secret": self._client_secret,
            }
        else:
            # Authenticate with a direct grant.
            data = {
                "grant_type": "password",
                "client_id": self._client_id,
                "client_secret": self._client_secret,
                "username": self._login,
                "password": self._password,
            }

        async with aiohttp.ClientSession() as session:
            async with session.post(self._token_url, data=data) as response:
                response.raise_for_status()
                ret = await response.json()
                await self._parse_response(ret)

    async def _refresh_tokens(self, refresh_token: str):
        await self._ensure_configuration()
        assert self._token_url is not None

        async with aiohttp.ClientSession() as session:
            async with session.post(
                self._token_url,
                data={
                    "grant_type": "refresh_token",
                    "client_id": self._client_id,
                    "client_secret": self._client_secret,
                    "refresh_token": refresh_token,
                },
            ) as response:
                if response.status == 401:
                    # Refresh token is invalid, get a new one.
                    await self._get_tokens()
                    return
                response.raise_for_status()
                ret = await response.json()
                await self._parse_response(ret)

    def clear_tokens(self):
        self._access_token = None
        self._decoded_access_token = None
        self._access_expired_at = None

        self._refresh_token = None
        self._decoded_refresh_token = None
        self._refresh_expired_at = None

    async def ensure_token(self):
        async with self._lock:
            now = datetime.now()
            if self._access_expired_at is None or self._access_expired_at <= now:
                refresh_token = self._refresh_token
                refresh_expired_at = self._refresh_expired_at
                self.clear_tokens()
                if refresh_expired_at is not None and refresh_expired_at > now:
                    assert refresh_token is not None
                    await self._refresh_tokens(refresh_token)
                else:
                    await self._get_tokens()

    async def get_token(self):
        await self.ensure_token()
        return self._access_token

    async def _run_with_auth(
        self,
        func: Callable[..., Coroutine[Any, Any, aiohttp.ClientResponse]],
        *args,
        **kwargs,
    ):
        kwargs["headers"] = kwargs.get("headers", {})
        kwargs["headers"]["Authorization"] = f"Bearer {await self.get_token()}"
        response = await func(*args, **kwargs)
        if response.status == 401:
            response.release()
            await response.wait_for_close()
            self.clear_tokens()
            kwargs["headers"]["Authorization"] = f"Bearer {await self.get_token()}"
            response = await func(*args, **kwargs)
        return response

    def with_auth(
        self,
        func: Callable[..., Coroutine[Any, Any, aiohttp.ClientResponse]],
        *args,
        **kwargs,
    ):
        return _AuthResponseContextManager(self._run_with_auth(func, *args, **kwargs))



import aiohttp
from pydantic import BaseModel


class ErrorInfo(BaseModel, extra="allow"):
    error: str
    message: str


class OzmaError(Exception):
    _info: ErrorInfo

    def __init__(self, info: ErrorInfo):
        self._info = info
        super().__init__(info.message)

    @property
    def info(self):
        return self._info


async def check_ozma_response(response: aiohttp.ClientResponse):
    if response.status == 200:
        return
    raw_error_info = await response.json()
    raise OzmaError(ErrorInfo.model_validate(raw_error_info))



from datetime import date, datetime, timedelta
from enum import StrEnum
from typing import Annotated, Any, Literal
import isodate
from pydantic import BaseModel, Field, PlainSerializer, model_validator
from pydantic.alias_generators import to_camel

from .render import ozmaql_name


DomainId = int
RowId = int
ArgumentName = str
InstanceName = str
InstanceName = str
SchemaName = str
EntityName = str
FieldName = str
ColumnName = str
IndexName = str
RoleName = str
UserViewName = str
ModuleName = str
ActionName = str
TriggerName = str
CustomEntityName = str
AttributeName = str


OzmaQLJson = dict[str, Any]
ScalarOzmaQLValue = str | int | float | bool | date | datetime | timedelta | OzmaQLJson | None
OzmaQLValue = ScalarOzmaQLValue | list[ScalarOzmaQLValue]
ScalarRawOzmaQLValue = str | int | float | bool | OzmaQLJson | None
RawOzmaQLValue = ScalarRawOzmaQLValue | list[ScalarRawOzmaQLValue]


def _ozmaql_scalar_value_to_raw(v: ScalarOzmaQLValue | ScalarRawOzmaQLValue) -> ScalarRawOzmaQLValue:
    match v:
        case date():
            return v.isoformat()
        case datetime():
            return v.isoformat()
        case timedelta():
            return isodate.duration_isoformat(v)
        case _:
            return v


def ozmaql_value_to_raw(v: OzmaQLValue | RawOzmaQLValue) -> RawOzmaQLValue:
    if isinstance(v, list):
        return [_ozmaql_scalar_value_to_raw(x) for x in v]
    else:
        return _ozmaql_scalar_value_to_raw(v)


def _ozmaql_scalar_value_from_raw(
    t: "OzmaDBScalarValueType", v: ScalarOzmaQLValue | ScalarRawOzmaQLValue
) -> ScalarOzmaQLValue:
    if v is None:
        return None
    match t:
        case OzmaDBDateValueType():
            match v:
                case date():
                    return v
                case str():
                    return date.fromisoformat(v)
                case _:
                    raise ValueError(f"Expected date string, got {v}")
        case OzmaDBDateTimeValueType() | OzmaDBLocalDateTimeValueType():
            match v:
                case datetime():
                    return v
                case str():
                    return datetime.fromisoformat(v)
                case _:
                    raise ValueError(f"Expected datetime string, got {v}")
        case OzmaDBIntervalValueType():
            match v:
                case timedelta():
                    return v
                case str():
                    return isodate.parse_duration(v)
                case _:
                    raise ValueError(f"Expected interval string, got {v}")
        case _:
            return v


def ozmaql_value_from_raw(t: "OzmaDBValueType", v: OzmaQLValue | RawOzmaQLValue) -> OzmaQLValue:
    if isinstance(t, OzmaDBArrayValueType):
        if not isinstance(v, list):
            raise ValueError(f"Expected list, got {v}")
        return [_ozmaql_scalar_value_from_raw(t.subtype, x) for x in v]
    else:
        if isinstance(v, list):
            raise ValueError(f"Expected scalar value, got {v}")
        return _ozmaql_scalar_value_from_raw(t, v)


OzmaQLValueArg = Annotated[
    OzmaQLValue | RawOzmaQLValue, PlainSerializer(ozmaql_value_to_raw, return_type=RawOzmaQLValue)
]


class OzmaDBModel(BaseModel, alias_generator=to_camel, populate_by_name=True, extra="ignore"):
    pass


class OzmaDBAbstractRef(BaseModel, alias_generator=to_camel, populate_by_name=True, frozen=True):
    # `schema` conflicts with `schema` in `BaseModel`.
    schema_name: SchemaName = Field(alias="schema")
    name: str

    # Allow positional args.
    def __init__(
        self, schema_name: SchemaName | None = None, name: str | None = None, schema: SchemaName | None = None
    ):
        if schema_name and schema:
            raise ValueError("Use schema_name or schema")
        super().__init__(schema_name=schema_name or schema, name=name)

    def __str__(self) -> str:
        return f"{ozmaql_name(self.schema_name)}.{ozmaql_name(self.name)}"


class OzmaDBEntityRef(OzmaDBAbstractRef, frozen=True):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class OzmaDBUserViewRef(OzmaDBAbstractRef, frozen=True):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        return f"&{ozmaql_name(self.schema_name)}.{ozmaql_name(self.name)}"


class OzmaDBFieldRef(BaseModel, alias_generator=to_camel, populate_by_name=True, frozen=True):
    entity: OzmaDBEntityRef
    name: str

    # Allow positional args.
    def __init__(self, entity: OzmaDBEntityRef, name: str):
        super().__init__(entity=entity, name=name)

    def __str__(self) -> str:
        return f"{self.entity}.{ozmaql_name(self.name)}"


class OzmaDBColumnField(OzmaDBModel):
    type: str = Field(description="The type of the column.")
    default_value: str | None = Field(default=None, description="Constant OzmaQL expression for the default value.")
    is_nullable: bool = False
    is_immutable: bool = False
    description: str = ""
    metadata: dict[str, Any] = Field(default_factory=dict)


class OzmaDBComputedField(OzmaDBModel):
    expression: str = Field(description="OzmaQL expression.")
    allow_broken: bool = False
    is_virtual: bool = False
    is_materialized: bool = False
    description: str = ""
    metadata: dict[str, Any] = Field(default_factory=dict)


class OzmaDBUniqueConstraint(OzmaDBModel):
    columns: list[str]
    is_alternate_key: bool = False
    description: str = ""
    metadata: dict[str, Any] = Field(default_factory=dict)


class OzmaDBCheckConstraint(OzmaDBModel):
    expression: str


class IndexType(StrEnum):
    BTREE = "btree"
    GIST = "gist"
    GIN = "gin"


class OzmaDBIndex(OzmaDBModel):
    expressions: list[str] = Field(description="OzmaQL expressions list to index.")
    included_expressions: list[str] = Field(default_factory=list)
    is_unique: bool = False
    type: IndexType = IndexType.BTREE
    predicate: str | None = None
    description: str = ""
    metadata: dict[str, Any] = Field(default_factory=dict)


class OzmaDBEntity(OzmaDBModel):
    column_fields: dict[FieldName, OzmaDBColumnField] = Field(default_factory=dict)
    computed_fields: dict[FieldName, OzmaDBComputedField] = Field(default_factory=dict)
    unique_constraints: dict[FieldName, OzmaDBUniqueConstraint] = Field(default_factory=dict)
    check_constraints: dict[FieldName, OzmaDBCheckConstraint] = Field(default_factory=dict)
    indexes: dict[IndexName, OzmaDBIndex] = Field(default_factory=dict)
    main_field: str | None = None
    save_restore_key: str | None = None
    is_abstract: bool = False
    is_frozen: bool = False
    parent: OzmaDBEntityRef | None = None
    description: str = ""
    metadata: dict[str, Any] = Field(default_factory=dict)


class OzmaDBAttributesField(OzmaDBModel):
    allow_broken: bool = False
    priority: int = 0
    attributes: str


class OzmaDBAttributesEntity(OzmaDBModel):
    fields: dict[FieldName, OzmaDBAttributesField] = Field(default_factory=dict)


class OzmaDBAttributesSchema(OzmaDBModel):
    entities: dict[EntityName, OzmaDBAttributesEntity] = Field(default_factory=dict)


class OzmaDBSavedSchemaCustomEntities(OzmaDBModel):
    type: Literal["customEntities"] = "customEntities"
    custom_entities: dict[SchemaName, dict[EntityName, list[Any]]] = Field(default_factory=dict)


class OzmaDBUserView(OzmaDBModel):
    query: str
    allow_broken: bool = False


class OzmaDBFullSavedSchema(OzmaDBModel):
    type: Literal["full"] = "full"
    custom_entities: dict[SchemaName, dict[EntityName, list[Any]]] = Field(default_factory=dict)
    entities: dict[EntityName, OzmaDBEntity] = Field(default_factory=dict)
    # TODO: Type others.
    roles: dict[RoleName, Any] = Field(default_factory=dict)
    user_views: dict[UserViewName, OzmaDBUserView] = Field(default_factory=dict)
    default_attributes: dict[SchemaName, OzmaDBAttributesSchema] = Field(default_factory=dict)
    modules: dict[ModuleName, Any] = Field(default_factory=dict)
    actions: dict[ActionName, Any] = Field(default_factory=dict)
    triggers: dict[TriggerName, Any] = Field(default_factory=dict)
    user_views_geneerator_script: str | None = None
    description: str = ""
    metadata: dict[str, Any] = Field(default_factory=dict)


OzmaDBSavedSchema = Annotated[OzmaDBFullSavedSchema | OzmaDBSavedSchemaCustomEntities, Field(discriminator="type")]


class OzmaDBSavedSchemas(OzmaDBModel):
    schemas: dict[SchemaName, OzmaDBSavedSchema] = Field(default_factory=dict)


class OzmaDBActionResponse(OzmaDBModel):
    result: RawOzmaQLValue = None


class OzmaDBInsertEntityOp(OzmaDBModel):
    type: Literal["insert"] = "insert"
    entity: OzmaDBEntityRef
    fields: dict[FieldName, OzmaQLValueArg]


class OzmaDBUpdateEntityOp(OzmaDBModel):
    type: Literal["update"] = "update"
    entity: OzmaDBEntityRef
    id: RowId
    fields: dict[FieldName, OzmaQLValueArg]


class OzmaDBDeleteEntityOp(OzmaDBModel):
    type: Literal["delete"] = "delete"
    entity: OzmaDBEntityRef
    id: RowId


OzmaDBTransactionOp = Annotated[
    OzmaDBInsertEntityOp | OzmaDBUpdateEntityOp | OzmaDBDeleteEntityOp, Field(discriminator="type")
]


class OzmaDBInsertEntityResult(OzmaDBModel):
    type: Literal["insert"] = "insert"
    id: RowId | None


class OzmaDBUpdateEntityResult(OzmaDBModel):
    type: Literal["update"] = "update"
    id: RowId


class OzmaDBDeleteEntityResult(OzmaDBModel):
    type: Literal["delete"] = "delete"


OzmaDBTransactionResultOp = Annotated[
    OzmaDBInsertEntityResult | OzmaDBUpdateEntityResult | OzmaDBDeleteEntityResult, Field(discriminator="type")
]


class OzmaDBTransactionResult(OzmaDBModel):
    results: list[OzmaDBTransactionResultOp]


class OzmaDBIntValueType(OzmaDBModel):
    type: Literal["int"] = "int"


class OzmaDBDecimalValueType(OzmaDBModel):
    type: Literal["decimal"] = "decimal"


class OzmaDBStringValueType(OzmaDBModel):
    type: Literal["string"] = "string"


class OzmaDBBoolValueType(OzmaDBModel):
    type: Literal["bool"] = "bool"


class OzmaDBDateTimeValueType(OzmaDBModel):
    type: Literal["datetime"] = "datetime"


class OzmaDBLocalDateTimeValueType(OzmaDBModel):
    type: Literal["localdatetime"] = "localdatetime"


class OzmaDBDateValueType(OzmaDBModel):
    type: Literal["date"] = "date"


class OzmaDBIntervalValueType(OzmaDBModel):
    type: Literal["interval"] = "interval"


class OzmaDBJsonValueType(OzmaDBModel):
    type: Literal["json"] = "json"


class OzmaDBUuidValueType(OzmaDBModel):
    type: Literal["uuid"] = "uuid"


OzmaDBScalarValueType = Annotated[
    OzmaDBIntValueType
    | OzmaDBDecimalValueType
    | OzmaDBStringValueType
    | OzmaDBBoolValueType
    | OzmaDBDateTimeValueType
    | OzmaDBLocalDateTimeValueType
    | OzmaDBDateValueType
    | OzmaDBIntervalValueType
    | OzmaDBJsonValueType
    | OzmaDBUuidValueType,
    Field(discriminator="type"),
]


class OzmaDBArrayValueType(OzmaDBModel):
    type: Literal["array"] = "array"
    subtype: OzmaDBScalarValueType


# OzmaDBValueType conversion
OzmaDBValueType = Annotated[OzmaDBScalarValueType | OzmaDBArrayValueType, Field(discriminator="type")]


class OzmaDBIntFieldType(OzmaDBModel):
    type: Literal["int"] = "int"


class OzmaDBDecimalFieldType(OzmaDBModel):
    type: Literal["decimal"] = "decimal"


class OzmaDBStringFieldType(OzmaDBModel):
    type: Literal["string"] = "string"


class OzmaDBBoolFieldType(OzmaDBModel):
    type: Literal["bool"] = "bool"


class OzmaDBDateTimeFieldType(OzmaDBModel):
    type: Literal["datetime"] = "datetime"


class OzmaDBDateFieldType(OzmaDBModel):
    type: Literal["date"] = "date"


class OzmaDBIntervalFieldType(OzmaDBModel):
    type: Literal["interval"] = "interval"


class OzmaDBJsonFieldType(OzmaDBModel):
    type: Literal["json"] = "json"


class OzmaDBUuidFieldType(OzmaDBModel):
    type: Literal["uuid"] = "uuid"


class OzmaDBReferenceFieldType(OzmaDBModel):
    type: Literal["reference"] = "reference"
    entity: OzmaDBEntityRef


class OzmaDBEnumFieldType(OzmaDBModel):
    type: Literal["enum"] = "enum"
    values: list[str]


OzmaDBScalarFieldType = Annotated[
    OzmaDBIntFieldType
    | OzmaDBDecimalFieldType
    | OzmaDBStringFieldType
    | OzmaDBBoolFieldType
    | OzmaDBDateTimeFieldType
    | OzmaDBDateFieldType
    | OzmaDBIntervalFieldType
    | OzmaDBJsonFieldType
    | OzmaDBUuidFieldType
    | OzmaDBReferenceFieldType
    | OzmaDBEnumFieldType,
    Field(discriminator="type"),
]


# ArrayFieldType
class OzmaDBArrayFieldType(OzmaDBModel):
    type: Literal["array"] = "array"
    subtype: OzmaDBScalarFieldType


# Union for FieldType
OzmaDBFieldType = Annotated[OzmaDBScalarFieldType | OzmaDBArrayFieldType, Field(discriminator="type")]


def ozmadb_scalar_field_type_to_type(type: OzmaDBScalarFieldType) -> OzmaDBScalarValueType:
    # Ugly!
    match type:
        case OzmaDBIntFieldType():
            return OzmaDBIntValueType()
        case OzmaDBDecimalFieldType():
            return OzmaDBDecimalValueType()
        case OzmaDBStringFieldType():
            return OzmaDBStringValueType()
        case OzmaDBBoolFieldType():
            return OzmaDBBoolValueType()
        case OzmaDBDateTimeFieldType():
            return OzmaDBDateTimeValueType()
        case OzmaDBDateFieldType():
            return OzmaDBDateValueType()
        case OzmaDBIntervalFieldType():
            return OzmaDBIntervalValueType()
        case OzmaDBJsonFieldType():
            return OzmaDBJsonValueType()
        case OzmaDBUuidFieldType():
            return OzmaDBUuidValueType()
        case OzmaDBReferenceFieldType():
            return OzmaDBIntValueType()
        case OzmaDBEnumFieldType():
            return OzmaDBStringValueType()
        case _:
            raise ValueError(f"Invalid scalar field type {type}")


def ozmadb_field_type_to_value(type: OzmaDBFieldType) -> OzmaDBValueType:
    if isinstance(type, OzmaDBArrayFieldType):
        return OzmaDBArrayValueType(subtype=ozmadb_scalar_field_type_to_type(type.subtype))
    else:
        return ozmadb_scalar_field_type_to_type(type)


class OzmaDBFieldAccess(OzmaDBModel):
    select: bool
    update: bool
    insert: bool


class OzmaDBViewColumnField(OzmaDBModel):
    field_type: OzmaDBFieldType
    value_type: OzmaDBValueType
    default_value: OzmaQLValue | None = None
    is_nullable: bool
    is_immutable: bool
    inherited_from: OzmaDBEntityRef | None = None
    access: OzmaDBFieldAccess
    hasUpdateTriggers: bool

    @model_validator(mode="after")
    def _validate(self: "OzmaDBViewColumnField") -> "OzmaDBViewColumnField":
        if self.default_value is not None:
            self.default_value = ozmaql_value_from_raw(self.value_type, self.default_value)
        return self


class OzmaDBMainFieldInfo(OzmaDBModel):
    name: FieldName
    field: OzmaDBViewColumnField


class OzmaDBBoundMappingEntry(OzmaDBModel):
    when: OzmaQLValue
    value: OzmaQLValue

    def _convert_values(self, type: OzmaDBValueType):
        self.when = ozmaql_value_from_raw(type, self.when)
        self.value = ozmaql_value_from_raw(type, self.value)


class OzmaDBBoundMapping(OzmaDBModel):
    entries: list[OzmaDBBoundMappingEntry]
    default: OzmaQLValue | None = None

    def _convert_values(self, type: OzmaDBValueType):
        for entry in self.entries:
            entry._convert_values(type)
        if self.default is not None:
            self.default = ozmaql_value_from_raw(type, self.default)


class OzmaDBBasicAttributeInfo(OzmaDBModel):
    type: OzmaDBValueType


class OzmaDBMappedAttributeInfo(OzmaDBBasicAttributeInfo):
    mapping: OzmaDBBoundMapping | None = None

    @model_validator(mode="after")
    def _validate(self: "OzmaDBMappedAttributeInfo") -> "OzmaDBMappedAttributeInfo":
        if self.mapping is not None:
            self.mapping._convert_values(self.type)
        return self


class OzmaDBBoundAttributeInfo(OzmaDBMappedAttributeInfo):
    const: bool = False


class OzmaDBCellAttributeInfo(OzmaDBMappedAttributeInfo):
    pass


class OzmaDBViewAttributeInfo(OzmaDBBasicAttributeInfo):
    const: bool = False


class OzmaDBRowAttributeInfo(OzmaDBBasicAttributeInfo):
    pass


class OzmaDBResultColumnInfo(OzmaDBModel):
    name: str
    attribute_types: dict[str, OzmaDBBoundAttributeInfo] = Field(default_factory=dict)
    cell_attribute_types: dict[str, OzmaDBCellAttributeInfo] = Field(default_factory=dict)
    value_type: OzmaDBValueType
    pun_type: OzmaDBValueType | None = None
    main_field: OzmaDBMainFieldInfo | None = None


class OzmaDBDomainField(OzmaDBModel):
    ref: OzmaDBFieldRef
    field: OzmaDBViewColumnField | None = None
    id_column: int


class OzmaDBArgument(OzmaDBModel):
    name: str
    arg_type: OzmaDBFieldType
    optional: bool
    default_value: OzmaQLValue | None = None
    attribute_types: dict[str, OzmaDBBoundAttributeInfo] = Field(default_factory=dict)

    @model_validator(mode="after")
    def _validate(self: "OzmaDBArgument") -> "OzmaDBArgument":
        arg_value_type = ozmadb_field_type_to_value(self.arg_type)
        if self.default_value is not None:
            self.default_value = ozmaql_value_from_raw(arg_value_type, self.default_value)
        return self


class OzmaDBMainEntity(OzmaDBModel):
    entity: OzmaDBEntityRef
    for_insert: bool


class OzmaDBResultViewInfo(OzmaDBModel):
    attribute_types: dict[str, OzmaDBViewAttributeInfo] = Field(default_factory=dict)
    row_attribute_types: dict[str, OzmaDBRowAttributeInfo] = Field(default_factory=dict)
    arguments: list[OzmaDBArgument]
    domains: dict[str, dict[str, OzmaDBDomainField]] = Field(default_factory=dict)
    main_entity: OzmaDBMainEntity | None = None
    columns: list[OzmaDBResultColumnInfo]
    hash: str


AttributesMap = dict[AttributeName, OzmaQLValue]


class OzmaDBExecutedValue(OzmaDBModel):
    value: OzmaQLValue
    attributes: AttributesMap | None = None
    pun: OzmaQLValue | None = None

    def _convert_values(self, column: OzmaDBResultColumnInfo):
        self.value = ozmaql_value_from_raw(column.value_type, self.value)
        if self.pun is not None:
            assert column.pun_type is not None
            self.pun = ozmaql_value_from_raw(column.pun_type, self.pun)
        if self.attributes is not None:
            for name, value in self.attributes.items():
                self.attributes[name] = ozmaql_value_from_raw(column.attribute_types[name].type, value)


class OzmaDBEntityId(OzmaDBModel):
    id: RowId
    sub_entity: OzmaDBEntityRef | None = None


class OzmaDBExecutedRow(OzmaDBModel):
    values: list[OzmaDBExecutedValue]
    domain_id: DomainId | None = None
    attributes: AttributesMap | None = None
    entity_ids: dict[ColumnName, OzmaDBEntityId] | None = None
    main_id: RowId | None = None
    main_sub_entity: OzmaDBEntityRef | None = None

    def _convert_values(self, info: OzmaDBResultViewInfo):
        for column, value in zip(info.columns, self.values):
            value._convert_values(column)
        if self.attributes is not None:
            for name, value in self.attributes.items():
                self.attributes[name] = ozmaql_value_from_raw(info.row_attribute_types[name].type, value)


class OzmaDBExecutedViewExpr(OzmaDBModel):
    attributes: AttributesMap
    column_attributes: list[AttributesMap]
    argument_attributes: dict[ArgumentName, AttributesMap]
    rows: list[OzmaDBExecutedRow]

    def _convert_values(self, info: OzmaDBResultViewInfo):
        for row in self.rows:
            row._convert_values(info)
        for name, value in self.attributes.items():
            self.attributes[name] = ozmaql_value_from_raw(info.attribute_types[name].type, value)
        for column, attributes in zip(info.columns, self.column_attributes):
            for name, value in attributes.items():
                attributes[name] = ozmaql_value_from_raw(column.attribute_types[name].type, value)
        for argument in info.arguments:
            attributes = self.argument_attributes.get(argument.name)
            if attributes is not None:
                for name, value in attributes.items():
                    attributes[name] = ozmaql_value_from_raw(argument.attribute_types[name].type, value)


class OzmaDBViewExprResult(OzmaDBModel):
    info: OzmaDBResultViewInfo
    result: OzmaDBExecutedViewExpr

    @model_validator(mode="after")
    def _validate(self: "OzmaDBViewExprResult") -> "OzmaDBViewExprResult":
        self.result._convert_values(self.info)
        return self


class OzmaDBViewInfoResult(OzmaDBModel):
    info: OzmaDBResultViewInfo
    const_attributes: AttributesMap
    const_column_attributes: list[AttributesMap]
    const_argument_attributes: dict[ArgumentName, AttributesMap]

    @model_validator(mode="after")
    def _validate(self: "OzmaDBViewInfoResult") -> "OzmaDBViewInfoResult":
        for name, value in self.const_attributes.items():
            self.const_attributes[name] = ozmaql_value_from_raw(self.info.attribute_types[name].type, value)
        for column, attributes in zip(self.info.columns, self.const_column_attributes):
            for name, value in attributes.items():
                attributes[name] = ozmaql_value_from_raw(column.attribute_types[name].type, value)
        for argument in self.info.arguments:
            attributes = self.const_argument_attributes.get(argument.name)
            if attributes is not None:
                for name, value in attributes.items():
                    attributes[name] = ozmaql_value_from_raw(argument.attribute_types[name].type, value)
        return self


class OzmaDBChunkArgument(OzmaDBModel):
    type: str
    value: OzmaQLValueArg


class OzmaDBChunkWhere(OzmaDBModel):
    arguments: dict[ArgumentName, OzmaDBChunkArgument] | None = None
    expression: str


class OzmaDBQueryChunk(OzmaDBModel):
    limit: int | None = None
    offset: int | None = None
    where: OzmaDBChunkWhere | None = None
    search: str | None = None
